import logging
from threading import Thread

import zmq
from zmq import Context
from zmq.devices import ThreadDevice

logger = logging.getLogger("ZeroMqEventManager.ForwarderDevice")


class ForwarderDevice(object):
    def __init__(self, context, pub_address, sub_address):
        # type: (Context, str, str) -> None
        self.context = context
        self.sub_address = pub_address
        self.push_address = sub_address

        self.frontend = None
        self.backend = None

        self.running = False
        self.device = None

    def start(self):
        # Socket facing clients
        self.device = ThreadDevice(zmq.FORWARDER, zmq.SUB, zmq.PUB)
        self.device.daemon = True
        self.device.setsockopt_in(zmq.SUBSCRIBE, b"")
        self.device.bind_in(self.sub_address)
        self.device.bind_out(self.push_address)
        self.device.start()

    def stop(self):
        pass



import logging
import logging.config
import os

import zmq
from diagnostics import DiagnosticMonitor
from helper import config_logger, read_json_from_file_as_dictionary
from messagehandlerbuilder import EventManagerMessageHandlerBuilder, ConfigKey
from zeromqmessagehandler import ZeroMqMessageBus, ZeroMqMessageHandler

COMP_NAME = "EventManager"

logger = logging.getLogger(COMP_NAME)


def main():
    config_path = './config.json'
    if "CONFIG_PATH" in os.environ:
        config_path = os.environ["CONFIG_PATH"]

    if os.path.isfile(config_path):
        config = read_json_from_file_as_dictionary(config_path)
    else:
        config = {}

    if "PUBLIC_HOSTNAME" in os.environ:
        config[ConfigKey.PUBLIC_HOSTNAME] = os.environ["PUBLIC_HOSTNAME"]

    log_path = config_path
    if "LOG_PATH" in os.environ:
        log_path = os.environ["LOG_PATH"]
    config_logger(log_path, COMP_NAME, add_error_handler=True)

    monitor = None
    if os.environ.get("DIAGNOSTIC_MONITOR", "").lower() in ("1", "true", "yes"):
        try:
            interval = int(os.environ.get("DIAGNOSTIC_INTERVAL", "1800"))
        except ValueError:
            interval = 1800
        monitor = DiagnosticMonitor(
            interval_seconds=interval,
            enable_tracemalloc=os.environ.get("DIAGNOSTIC_TRACEMALLOC", "").lower() in ("1", "true"),
        )
        monitor.start()

    context = zmq.Context()

    message_bus = ZeroMqMessageBus(COMP_NAME, context)

    message_handler_builder = EventManagerMessageHandlerBuilder(context, config, message_bus, logger)
    message_handler = ZeroMqMessageHandler(context,
                                           COMP_NAME,
                                           message_bus,
                                           message_handler_builder)
    try:
        message_handler.handle_messages()
    finally:
        if monitor is not None:
            monitor.stop()
        message_handler_builder.destroy_singletons()
    message_bus.dispose()


if __name__ == '__main__':
    main()



import gc
import json
import logging
import os
import threading
import time
import tracemalloc
from collections import defaultdict
from typing import Any, Dict, List, Optional, Set, Tuple


TCP_STATES = {
    "01": "ESTABLISHED",
    "02": "SYN_SENT",
    "03": "SYN_RECV",
    "04": "FIN_WAIT1",
    "05": "FIN_WAIT2",
    "06": "TIME_WAIT",
    "07": "CLOSE",
    "08": "CLOSE_WAIT",
    "09": "LAST_ACK",
    "0A": "LISTEN",
    "0B": "CLOSING",
}


def _parse_ipv4_addr(hex_addr: str) -> Tuple[str, int]:
    addr_hex, port_hex = hex_addr.split(":")
    port = int(port_hex, 16)
    addr_int = int(addr_hex, 16)
    ip = "{}.{}.{}.{}".format(
        addr_int & 0xFF,
        (addr_int >> 8) & 0xFF,
        (addr_int >> 16) & 0xFF,
        (addr_int >> 24) & 0xFF,
    )
    return ip, port


def _parse_ipv6_addr(hex_addr: str) -> Tuple[str, int]:
    addr_hex, port_hex = hex_addr.split(":")
    port = int(port_hex, 16)

    if len(addr_hex) != 32:
        return addr_hex, port

    groups = [addr_hex[i:i + 8] for i in range(0, 32, 8)]
    reversed_groups = []
    for g in groups:
        b = bytes.fromhex(g)
        reversed_groups.append(b[::-1].hex())

    full_hex = "".join(reversed_groups)

    if full_hex == "00000000000000000000ffff" + full_hex[24:]:
        ipv4_hex = full_hex[24:]
        octets = [int(ipv4_hex[i:i + 2], 16) for i in range(0, 8, 2)]
        ip = "::ffff:{}.{}.{}.{}".format(*octets)
    elif full_hex == "0" * 32:
        ip = "::"
    elif full_hex == "0" * 31 + "1":
        ip = "::1"
    else:
        ipv6_parts = [full_hex[i:i + 4] for i in range(0, 32, 4)]
        ip = ":".join(ipv6_parts)

    return ip, port


class DiagnosticMonitor:
    def __init__(
        self,
        interval_seconds: int = 1800,
        enable_tracemalloc: bool = False,
        tracemalloc_top_n: int = 15,
    ) -> None:
        self._interval = interval_seconds
        self._enable_tracemalloc = enable_tracemalloc
        self._tracemalloc_top_n = tracemalloc_top_n

        self._running = False
        self._thread: Optional[threading.Thread] = None
        self._seq = 0
        self._start_time = 0.0
        self._prev: Optional[Dict[str, Any]] = None
        self._prev_tracemalloc = None
        self._prev_thread_names: Set[str] = set()
        self._cumulative_thread_churn = 0

        self._logger = logging.getLogger("EventManagerDiagnosticMonitor")

        if self._enable_tracemalloc and not tracemalloc.is_tracing():
            tracemalloc.start(25)

    def start(self) -> None:
        if self._running:
            return
        self._running = True
        self._start_time = time.time()
        self._thread = threading.Thread(
            target=self._loop,
            name="DiagnosticMonitor",
            daemon=True,
        )
        self._thread.start()
        self._logger.warning("[DIAG] Monitor started - interval=%ds tracemalloc=%s", self._interval, self._enable_tracemalloc)

    def stop(self) -> None:
        self._running = False
        if self._thread is not None:
            self._thread.join(timeout=10)

    def _loop(self) -> None:
        self._wait(60)
        while self._running:
            try:
                self._snapshot()
            except Exception:
                self._logger.exception("[DIAG] Snapshot failed")
            self._wait(self._interval)

    def _wait(self, seconds: int) -> None:
        end = time.time() + seconds
        while self._running and time.time() < end:
            time.sleep(1)

    def _snapshot(self) -> None:
        self._seq += 1
        uptime_h = round((time.time() - self._start_time) / 3600, 2)

        snap: Dict[str, Any] = {
            "seq": self._seq,
            "uptime_h": uptime_h,
            "ts": time.strftime("%Y-%m-%dT%H:%M:%S%z"),
        }

        snap["mem"] = self._collect_memory()
        snap["fd"] = self._collect_fd()
        snap["tcp"] = self._collect_tcp()
        snap["threads"] = self._collect_threads()
        snap["locks"] = self._collect_locks()
        snap["zmq_objects"] = self._collect_zmq_objects()
        snap["gc"] = self._collect_gc()

        if self._enable_tracemalloc:
            snap["tracemalloc"] = self._collect_tracemalloc()

        if self._prev is not None:
            snap["delta"] = self._compute_delta(snap, self._prev)

        self._log(snap)

        snap.get("tcp", {}).pop("details", None)
        self._prev = snap

    # -- collectors --

    def _collect_memory(self) -> Dict[str, Any]:
        result: Dict[str, Any] = {}
        try:
            with open("/proc/self/status", "r") as f:
                for line in f:
                    key = line.split(":")[0]
                    if key in ("VmRSS", "VmSize", "VmPeak", "VmHWM", "RssAnon", "RssFile"):
                        result[key] = int(line.split()[1])
        except (FileNotFoundError, PermissionError):
            result["error"] = "proc not available"
        return result

    def _collect_fd(self) -> Dict[str, Any]:
        result: Dict[str, Any] = {"total": 0, "by_type": {}}
        try:
            fd_dir = "/proc/self/fd"
            fds = os.listdir(fd_dir)
            result["total"] = len(fds)

            counts: Dict[str, int] = defaultdict(int)
            files: List[str] = []

            for fd in fds:
                try:
                    target = os.readlink(os.path.join(fd_dir, fd))
                    if target.startswith("socket:"):
                        counts["socket"] += 1
                    elif target.startswith("pipe:"):
                        counts["pipe"] += 1
                    elif "eventfd" in target:
                        counts["eventfd"] += 1
                    elif "eventpoll" in target:
                        counts["eventpoll"] += 1
                    elif target.startswith("anon_inode:"):
                        counts["anon_inode"] += 1
                    else:
                        counts["file"] += 1
                        if len(files) < 20:
                            files.append(target)
                except OSError:
                    counts["unknown"] += 1

            result["by_type"] = dict(counts)
            if files:
                result["files"] = files
        except (FileNotFoundError, PermissionError):
            result["error"] = "proc not available"
        return result

    def _collect_tcp(self) -> Dict[str, Any]:
        by_state: Dict[str, int] = defaultdict(int)
        details: List[Dict[str, Any]] = []
        total_tx = 0
        total_rx = 0

        for tcp_path in ("/proc/self/net/tcp", "/proc/self/net/tcp6"):
            is_v6 = "tcp6" in tcp_path
            try:
                with open(tcp_path, "r") as f:
                    for line in f.readlines()[1:]:
                        parts = line.strip().split()
                        if len(parts) < 10:
                            continue

                        state = TCP_STATES.get(parts[3], "UNKNOWN({})".format(parts[3]))
                        by_state[state] += 1

                        tx_hex, rx_hex = parts[4].split(":")
                        tx = int(tx_hex, 16)
                        rx = int(rx_hex, 16)
                        total_tx += tx
                        total_rx += rx

                        parse = _parse_ipv6_addr if is_v6 else _parse_ipv4_addr
                        local_ip, local_port = parse(parts[1])
                        remote_ip, remote_port = parse(parts[2])

                        entry: Dict[str, Any] = {
                            "local": "{}:{}".format(local_ip, local_port),
                            "remote": "{}:{}".format(remote_ip, remote_port),
                            "state": state,
                            "inode": parts[9],
                        }
                        if tx > 0:
                            entry["tx"] = tx
                        if rx > 0:
                            entry["rx"] = rx

                        details.append(entry)
            except (FileNotFoundError, PermissionError):
                continue

        return {
            "by_state": dict(by_state),
            "details": details,
            "total_tx": total_tx,
            "total_rx": total_rx,
        }

    def _collect_threads(self) -> Dict[str, Any]:
        threads = threading.enumerate()
        groups: Dict[str, int] = defaultdict(int)
        names: List[str] = []

        for t in threads:
            name = t.name
            names.append(name)
            prefix = name.split("_Worker_")[0] if "_Worker_" in name else name
            groups[prefix] += 1

        current_names = set(names)
        new_threads = current_names - self._prev_thread_names
        self._cumulative_thread_churn += len(new_threads)
        self._prev_thread_names = current_names

        return {
            "total": len(threads),
            "by_group": dict(groups),
            "new_since_last": list(new_threads) if self._seq > 1 else [],
            "cumulative_churn": self._cumulative_thread_churn,
        }

    def _collect_locks(self) -> Dict[str, Any]:
        result: Dict[str, Any] = {"total_contexts": 0, "total_locks": 0, "top_contexts": []}
        try:
            from zeromqmessagehandler import HandleEventWorkerThread
            lock = HandleEventWorkerThread.non_reentrant_lock
            locks_dict = HandleEventWorkerThread.non_reentrant_context_locks

            with lock:
                context_sizes = {str(ctx): len(subjects) for ctx, subjects in locks_dict.items()}

            result["total_contexts"] = len(context_sizes)
            result["total_locks"] = sum(context_sizes.values())

            sorted_ctx = sorted(context_sizes.items(), key=lambda x: x[1], reverse=True)
            result["top_contexts"] = sorted_ctx[:10]
        except Exception as e:
            result["error"] = str(e)
        return result

    def _collect_zmq_objects(self) -> Dict[str, Any]:
        zmq_socket_count = 0
        socket_resource_count = 0
        lock_count = 0

        try:
            all_objects = gc.get_objects()
        except MemoryError:
            return {"error": "MemoryError - heap too large for gc.get_objects()"}

        try:
            for obj in all_objects:
                type_name = type(obj).__name__
                if type_name == "Socket":
                    zmq_socket_count += 1
                elif type_name == "SocketResource":
                    socket_resource_count += 1
                elif type_name in ("lock", "_RLock"):
                    lock_count += 1
        finally:
            del all_objects

        return {
            "zmq_socket": zmq_socket_count,
            "socket_resource": socket_resource_count,
            "lock": lock_count,
        }

    def _collect_gc(self) -> Dict[str, Any]:
        stats = gc.get_stats()
        return {
            "garbage": len(gc.garbage),
            "generations": [
                {"coll": s["collections"], "ok": s["collected"], "fail": s["uncollectable"]}
                for s in stats
            ],
        }

    def _collect_tracemalloc(self) -> Dict[str, Any]:
        if not tracemalloc.is_tracing():
            return {"enabled": False}

        snapshot = tracemalloc.take_snapshot()
        top = snapshot.statistics("lineno")

        result: Dict[str, Any] = {
            "enabled": True,
            "total_kb": round(sum(s.size for s in top) / 1024, 1),
            "top": [
                {"loc": str(s.traceback), "kb": round(s.size / 1024, 1), "n": s.count}
                for s in top[:self._tracemalloc_top_n]
            ],
        }

        if self._prev_tracemalloc is not None:
            diff = snapshot.compare_to(self._prev_tracemalloc, "lineno")
            result["growth"] = [
                {"loc": str(s.traceback), "diff_kb": round(s.size_diff / 1024, 1), "n": s.count_diff}
                for s in diff[:10]
                if s.size_diff > 0
            ]

        self._prev_tracemalloc = snapshot
        return result

    # -- delta --

    def _compute_delta(self, curr: Dict[str, Any], prev: Dict[str, Any]) -> Dict[str, Any]:
        d: Dict[str, Any] = {}

        d["fd"] = curr.get("fd", {}).get("total", 0) - prev.get("fd", {}).get("total", 0)
        d["rss_kb"] = curr.get("mem", {}).get("VmRSS", 0) - prev.get("mem", {}).get("VmRSS", 0)
        d["threads"] = curr.get("threads", {}).get("total", 0) - prev.get("threads", {}).get("total", 0)
        d["locks"] = curr.get("locks", {}).get("total_locks", 0) - prev.get("locks", {}).get("total_locks", 0)
        d["zmq_socket"] = curr.get("zmq_objects", {}).get("zmq_socket", 0) - prev.get("zmq_objects", {}).get("zmq_socket", 0)

        curr_states = curr.get("tcp", {}).get("by_state", {})
        prev_states = prev.get("tcp", {}).get("by_state", {})
        state_delta = {}
        for s in set(list(curr_states.keys()) + list(prev_states.keys())):
            diff = curr_states.get(s, 0) - prev_states.get(s, 0)
            if diff != 0:
                state_delta[s] = diff
        if state_delta:
            d["tcp_states"] = state_delta

        fd_curr = curr.get("fd", {}).get("by_type", {})
        fd_prev = prev.get("fd", {}).get("by_type", {})
        fd_delta = {}
        for t in set(list(fd_curr.keys()) + list(fd_prev.keys())):
            diff = fd_curr.get(t, 0) - fd_prev.get(t, 0)
            if diff != 0:
                fd_delta[t] = diff
        if fd_delta:
            d["fd_by_type"] = fd_delta

        return d

    # -- logging --

    def _log(self, snap: Dict[str, Any]) -> None:
        seq = snap["seq"]
        up = snap["uptime_h"]
        mem = snap.get("mem", {})
        fd = snap.get("fd", {})
        tcp = snap.get("tcp", {})
        thr = snap.get("threads", {})
        locks = snap.get("locks", {})
        zmq_obj = snap.get("zmq_objects", {})

        rss_mb = mem.get("VmRSS", 0) / 1024
        vms_mb = mem.get("VmSize", 0) / 1024
        peak_mb = mem.get("VmHWM", 0) / 1024

        self._logger.warning(
            "[DIAG] seq=%d uptime=%.1fh rss=%.1fMB(peak=%.1fMB) vms=%.1fMB "
            "fd=%d(sock=%d,pipe=%d,evfd=%d) threads=%d(churn=%d) "
            "locks=%d/%dctx zmq_sock=%d sock_res=%d",
            seq, up, rss_mb, peak_mb, vms_mb,
            fd.get("total", 0),
            fd.get("by_type", {}).get("socket", 0),
            fd.get("by_type", {}).get("pipe", 0),
            fd.get("by_type", {}).get("eventfd", 0),
            thr.get("total", 0),
            thr.get("cumulative_churn", 0),
            locks.get("total_locks", 0),
            locks.get("total_contexts", 0),
            zmq_obj.get("zmq_socket", 0),
            zmq_obj.get("socket_resource", 0),
        )

        self._log_tcp_sockets(seq, tcp)
        self._log_thread_groups(seq, thr)
        self._log_lock_top(seq, locks)
        self._log_delta(seq, snap.get("delta"))
        self._log_gc(seq, snap.get("gc", {}))

        if "tracemalloc" in snap:
            self._log_tracemalloc(seq, snap["tracemalloc"])

        json_snap = dict(snap)
        tcp_copy = dict(json_snap.get("tcp", {}))
        tcp_copy.pop("details", None)
        json_snap["tcp"] = tcp_copy
        self._logger.info("[DIAG:JSON] seq=%d %s", seq, json.dumps(json_snap, default=str, separators=(",", ":")))

    def _log_tcp_sockets(self, seq: int, tcp: Dict[str, Any]) -> None:
        by_state = tcp.get("by_state", {})
        if not by_state:
            return

        state_str = " ".join("{}={}".format(k, v) for k, v in sorted(by_state.items()))
        self._logger.warning(
            "[DIAG:TCP] seq=%d %s tx_queue=%dB rx_queue=%dB",
            seq, state_str, tcp.get("total_tx", 0), tcp.get("total_rx", 0),
        )

        details = tcp.get("details", [])

        listen = [d["local"] for d in details if d["state"] == "LISTEN"]
        if listen:
            self._logger.warning("[DIAG:LISTEN] seq=%d %s", seq, " ".join(sorted(set(listen))))

        conn_groups: Dict[str, int] = defaultdict(int)
        for d in details:
            if d["state"] == "LISTEN":
                continue
            key = "->{}({})".format(d["remote"], d["state"][:3])
            conn_groups[key] += 1

        if conn_groups:
            sorted_conns = sorted(conn_groups.items(), key=lambda x: -x[1])
            parts = ["{}x{}".format(k, v) for k, v in sorted_conns[:30]]
            self._logger.warning("[DIAG:CONN] seq=%d %s", seq, " ".join(parts))

    def _log_thread_groups(self, seq: int, thr: Dict[str, Any]) -> None:
        groups = thr.get("by_group", {})
        if groups:
            group_str = " ".join("{}={}".format(k, v) for k, v in sorted(groups.items()))
            self._logger.warning("[DIAG:THREADS] seq=%d %s", seq, group_str)

        new = thr.get("new_since_last", [])
        if new:
            self._logger.warning("[DIAG:NEW_THREADS] seq=%d count=%d names=%s", seq, len(new), ",".join(sorted(new)[:20]))

    def _log_lock_top(self, seq: int, locks: Dict[str, Any]) -> None:
        top = locks.get("top_contexts", [])
        if not top:
            return
        parts = ["{}={}".format(ctx, count) for ctx, count in top[:5]]
        self._logger.warning("[DIAG:LOCKS_TOP] seq=%d %s", seq, " ".join(parts))

    def _log_delta(self, seq: int, delta: Optional[Dict[str, Any]]) -> None:
        if delta is None:
            return

        simple = []
        for key in ("fd", "rss_kb", "threads", "locks", "zmq_socket"):
            val = delta.get(key, 0)
            if val != 0:
                simple.append("{}={:+d}".format(key, val))

        tcp_d = delta.get("tcp_states", {})
        for state, val in sorted(tcp_d.items()):
            simple.append("{}={:+d}".format(state, val))

        fd_d = delta.get("fd_by_type", {})
        for fd_type, val in sorted(fd_d.items()):
            simple.append("fd_{}={:+d}".format(fd_type, val))

        if simple:
            self._logger.warning("[DIAG:DELTA] seq=%d %s", seq, " ".join(simple))

    def _log_gc(self, seq: int, gc_info: Dict[str, Any]) -> None:
        garbage = gc_info.get("garbage", 0)
        gens = gc_info.get("generations", [])
        if not gens:
            return

        gen_parts = []
        for i, g in enumerate(gens):
            gen_parts.append("gen{}={}/{}/{}".format(i, g["coll"], g["ok"], g["fail"]))

        self._logger.info("[DIAG:GC] seq=%d garbage=%d %s", seq, garbage, " ".join(gen_parts))

    def _log_tracemalloc(self, seq: int, tm: Dict[str, Any]) -> None:
        if not tm.get("enabled"):
            return

        self._logger.warning("[DIAG:TRACEMALLOC] seq=%d total=%.1fKB", seq, tm.get("total_kb", 0))

        for alloc in tm.get("top", [])[:5]:
            self._logger.info(
                "[DIAG:TRACEMALLOC_TOP] seq=%d %.1fKB n=%d %s",
                seq, alloc["kb"], alloc["n"], alloc["loc"],
            )

        for g in tm.get("growth", [])[:5]:
            self._logger.warning(
                "[DIAG:TRACEMALLOC_GROWTH] seq=%d +%.1fKB n=%+d %s",
                seq, g["diff_kb"], g["n"], g["loc"],
            )



from logging import Logger

from event_manager_api.shared.token import TK_EVENT_MANAGER_SUBSCRIBE
from messagehandler import MessageHandlerBuilder
from messageprocessor import MessageProcessorMessageHandler, DefaultMessageProcessorExecutorFactory
from messageprocessorutil import LoggingProcessorCallback
from zmq import Context

from zeromqeventmanager import (
    TK_EVENT_MANAGER_UNSUBSCRIBE,
    TK_EVENT_MANAGER_PUBLISH,
    TK_EVENT_MANAGER_KEEP_ALIVE,
    EventManager,
)
from zeromqeventmanager.processor import (
    SubscribeProcessor,
    PublishProcessor,
    KeepAliveProcessor,
)
from zeromqmessagehandler import ZeroMqMessageBus

from ._ConfigKey import ConfigKey


class EventManagerMessageHandlerBuilder(MessageHandlerBuilder):
    def __init__(self, context, config, message_bus, logger):
        # type: (Context, dict, ZeroMqMessageBus, Logger) -> None
        self.context = context
        self.config = config
        self.message_bus = message_bus
        self.logger = logger
        self.event_manager = None  # type: EventManager

    def build_singletons(self):
        self.event_manager = EventManager(
            context=self.context,
            message_bus=self.message_bus,
            pub_address=self.config.get(ConfigKey.PUB_ADDRESS),
            sub_address=self.config.get(ConfigKey.SUB_ADDRESS),
            pull_address=self.config.get(ConfigKey.PULL_ADDRESS),
            public_hostname=self.config.get(ConfigKey.PUBLIC_HOSTNAME),
            min_port=self.config.get(ConfigKey.MIN_STREAMER_PORT, 5001),
            max_port=self.config.get(ConfigKey.MAX_STREAMER_PORT, 5100),
            clean_interval=5,
            sub_id_lifetime=30,
            streamer_poll_timeout=self.config.get(ConfigKey.STREAMER_POLL_TIMEOUT, 1000)
        )

    def build_message_handler(self, message, event):
        handled_tokens = {
            TK_EVENT_MANAGER_SUBSCRIBE: SubscribeProcessor(self.event_manager),
            TK_EVENT_MANAGER_UNSUBSCRIBE: None,
            TK_EVENT_MANAGER_PUBLISH: PublishProcessor(self.event_manager),
            TK_EVENT_MANAGER_KEEP_ALIVE: KeepAliveProcessor(self.event_manager)
        }
        return MessageProcessorMessageHandler(
            None,
            handled_tokens,
            DefaultMessageProcessorExecutorFactory([LoggingProcessorCallback(self.logger)]))

    def destroy_message_handler(self, message_handler):
        pass

    def destroy_singletons(self):
        if self.event_manager is not None:
            self.event_manager.stop()



import json
import unittest
import time
from datetime import datetime, timedelta
from threading import Thread

import zmq
from messagebus import Event, Message, TokenCreator, TokenPriority, DataType, DefaultToken
from zeromqmessagehandler import ZeroMqMessageBus

SERVICE_LOCATOR_GROUP = "3"
TK_SERVICE_LOCATOR_LOOKUP_SERVICE = TokenCreator.create_token(TokenPriority.high, SERVICE_LOCATOR_GROUP, "2")


class PublishTestCase(unittest.TestCase):
    def test_publish(self):
        ctx = zmq.Context()

        pub_socket = ctx.socket(zmq.PUB)
        pub_socket.connect("tcp://localhost:5000")

        data1 = {

        }
        data2 = {
            "id": "asdf",
            "code": "asdf",
            "partner": "Teste",
            "storeCode": "99994"
        }
        data3 = {
            "orderCode": "12",
            "partnerId": "8",
            "fromStatus": "",
            "toStatus": ""
        }
        data4 = {
        }
        data5 = {
            "id": "asdf"
        }
        event1 = Event("StoreStatusUpdate", data=json.dumps(data1).encode("utf-8"))
        event2 = Event("OrderManagerOrderIntegrated", data=json.dumps(data2).encode("utf-8"))
        event3 = Event("OrderStatusCodeUpdated", data=json.dumps(data3).encode("utf-8"))
        event4 = Event("PublishMenuVersion", data=json.dumps(data4).encode("utf-8"), headers={"context": "localhost"})
        event5 = Event("PosOrderConfirm", data=json.dumps(data5).encode("utf-8"))
        method = "POST"
        path = "/mwboh-api/v1.0.0/sale/order-picture"
        headers = {
            "x-api-key": "58b333b4-449b-4f08-b81a-bd48b658e266"
        }
        data = f"{method}\0{path}\0{len(headers)}\0"
        if len(headers) > 0:
            for header in headers.keys():
                data += header + "\0" + headers[header]
            data += "\0"
        query_params = {}
        data += str(len(query_params)) + "\0"
        if len(query_params) > 0:
            for query_param in query_params.keys():
                data += query_param + "\0" + query_params[query_param]
            data += "\0"
        body = """{"order-code":"9","order-picture-sale-lines":[{"comment":[],"unit-price":4.9,"menu-item-code":"1","qty":1,"inc-qty":0,"sub-unit-price":0.0,"tax-items":[{"base-amount-bd":4.9,"base-amount-ad":2.29,"tax-index":"PIS;P","tax-amount-bd":0.196,"tax-rule-id":1955574606,"tax-rate":0.04,"tax-amount-ad":0.0916},{"base-amount-bd":4.9,"base-amount-ad":2.29,"tax-index":"COFINS;C","tax-amount-bd":0.245,"tax-rule-id":2143042763,"tax-rate":0.05,"tax-amount-ad":0.1145},{"base-amount-bd":4.9,"base-amount-ad":2.29,"tax-index":"ICMS;I","tax-amount-bd":0.637,"tax-rule-id":1707847563,"tax-rate":0.13,"tax-amount-ad":0.2977}],"default-qty":null,"item-type":"PRODUCT","sub-qty":null,"discount-applied":"600:2.61","added-qty":null,"product":"X-Burger","item-discount":2.61,"dec-qty":0,"multiplied-qty":1,"chosen-qty":null,"price-key":"249108103194","added-unit-price":0.0,"level":0,"item-price":4.9,"order-picture-id":6,"custom-properties":{"a":"b"},"part-code":"1001","plu":"1001","line-number":1},{"comment":[],"unit-price":0.0,"menu-item-code":"1.1001","qty":0,"inc-qty":0,"sub-unit-price":0.0,"tax-items":[],"default-qty":0,"item-type":"OPTION","sub-qty":0,"discount-applied":null,"added-qty":0,"product":"Ponto da Carne","item-discount":0.0,"dec-qty":0,"multiplied-qty":0.0,"chosen-qty":0,"price-key":null,"added-unit-price":0.0,"level":1,"item-price":0.0,"order-picture-id":6,"custom-properties":{},"part-code":"700000001","plu":"700000001","line-number":1},{"comment":[],"unit-price":0.0,"menu-item-code":"1.1001","qty":1,"inc-qty":0,"sub-unit-price":0.0,"tax-items":[],"default-qty":1,"item-type":"OPTION","sub-qty":0,"discount-applied":null,"added-qty":0,"product":"Maionese Extra","item-discount":0.0,"dec-qty":0,"multiplied-qty":1.0,"chosen-qty":1,"price-key":null,"added-unit-price":0.0,"level":1,"item-price":0.0,"order-picture-id":6,"custom-properties":{},"part-code":"900000002","plu":"900000002","line-number":1},{"comment":[],"unit-price":0.0,"menu-item-code":"1.1001.900000002","qty":1,"inc-qty":0,"sub-unit-price":0.0,"tax-items":[{"base-amount-bd":2.0,"base-amount-ad":0.94,"tax-index":"PIS;P","tax-amount-bd":0.07,"tax-rule-id":27939140,"tax-rate":0.035,"tax-amount-ad":0.0329},{"base-amount-bd":2.0,"base-amount-ad":0.94,"tax-index":"COFINS;C","tax-amount-bd":0.07,"tax-rule-id":190559478,"tax-rate":0.035,"tax-amount-ad":0.0329},{"base-amount-bd":2.0,"base-amount-ad":0.94,"tax-index":"ICMS;I","tax-amount-bd":0.24,"tax-rule-id":1678362044,"tax-rate":0.12,"tax-amount-ad":0.1128}],"default-qty":null,"item-type":"CANADD","sub-qty":null,"discount-applied":"600:1.06","added-qty":null,"product":"Maionese","item-discount":1.06,"dec-qty":0,"multiplied-qty":1.0,"chosen-qty":null,"price-key":"249108103236","added-unit-price":2.0,"level":2,"item-price":2.0,"order-picture-id":6,"custom-properties":{},"part-code":"10101","plu":"10101","line-number":1},{"comment":[],"unit-price":0.0,"menu-item-code":"1","qty":1,"inc-qty":0,"sub-unit-price":0.0,"tax-items":[],"default-qty":null,"item-type":"COMBO","sub-qty":null,"discount-applied":null,"added-qty":null,"product":"Combo X-Burger","item-discount":0.0,"dec-qty":0,"multiplied-qty":1,"chosen-qty":null,"price-key":null,"added-unit-price":0.0,"level":0,"item-price":0.0,"order-picture-id":6,"custom-properties":{},"part-code":"20001","plu":"20001","line-number":2},{"comment":[],"unit-price":4.9,"menu-item-code":"1.20001","qty":1,"inc-qty":0,"sub-unit-price":0.0,"tax-items":[{"base-amount-bd":4.9,"base-amount-ad":2.29,"tax-index":"PIS;P","tax-amount-bd":0.196,"tax-rule-id":1955574606,"tax-rate":0.04,"tax-amount-ad":0.0916},{"base-amount-bd":4.9,"base-amount-ad":2.29,"tax-index":"COFINS;C","tax-amount-bd":0.245,"tax-rule-id":2143042763,"tax-rate":0.05,"tax-amount-ad":0.1145},{"base-amount-bd":4.9,"base-amount-ad":2.29,"tax-index":"ICMS;I","tax-amount-bd":0.637,"tax-rule-id":1707847563,"tax-rate":0.13,"tax-amount-ad":0.2977}],"default-qty":1,"item-type":"PRODUCT","sub-qty":0,"discount-applied":"600:2.61","added-qty":0,"product":"X-Burger","item-discount":2.61,"dec-qty":0,"multiplied-qty":1.0,"chosen-qty":null,"price-key":"249108103219","added-unit-price":0.0,"level":1,"item-price":4.9,"order-picture-id":6,"custom-properties":{},"part-code":"1001","plu":"1001","line-number":2},{"comment":[],"unit-price":0.0,"menu-item-code":"1.20001.1001","qty":0,"inc-qty":0,"sub-unit-price":0.0,"tax-items":[],"default-qty":0,"item-type":"OPTION","sub-qty":0,"discount-applied":null,"added-qty":0,"product":"Ponto da Carne","item-discount":0.0,"dec-qty":0,"multiplied-qty":0.0,"chosen-qty":0,"price-key":null,"added-unit-price":0.0,"level":2,"item-price":0.0,"order-picture-id":6,"custom-properties":{},"part-code":"700000001","plu":"700000001","line-number":2},{"comment":[],"unit-price":0.0,"menu-item-code":"1.20001.1001","qty":1,"inc-qty":0,"sub-unit-price":0.0,"tax-items":[],"default-qty":1,"item-type":"OPTION","sub-qty":0,"discount-applied":null,"added-qty":0,"product":"Maionese Extra","item-discount":0.0,"dec-qty":0,"multiplied-qty":1.0,"chosen-qty":1,"price-key":null,"added-unit-price":0.0,"level":2,"item-price":0.0,"order-picture-id":6,"custom-properties":{},"part-code":"900000002","plu":"900000002","line-number":2},{"comment":[],"unit-price":0.0,"menu-item-code":"1.20001.1001.900000002","qty":1,"inc-qty":0,"sub-unit-price":0.0,"tax-items":[{"base-amount-bd":2.0,"base-amount-ad":0.94,"tax-index":"PIS;P","tax-amount-bd":0.07,"tax-rule-id":27939140,"tax-rate":0.035,"tax-amount-ad":0.0329},{"base-amount-bd":2.0,"base-amount-ad":0.94,"tax-index":"COFINS;C","tax-amount-bd":0.07,"tax-rule-id":190559478,"tax-rate":0.035,"tax-amount-ad":0.0329},{"base-amount-bd":2.0,"base-amount-ad":0.94,"tax-index":"ICMS;I","tax-amount-bd":0.24,"tax-rule-id":1678362044,"tax-rate":0.12,"tax-amount-ad":0.1128}],"default-qty":null,"item-type":"CANADD","sub-qty":null,"discount-applied":"600:1.06","added-qty":null,"product":"Maionese","item-discount":1.06,"dec-qty":0,"multiplied-qty":1.0,"chosen-qty":null,"price-key":"249108103236","added-unit-price":2.0,"level":3,"item-price":2.0,"order-picture-id":6,"custom-properties":{},"part-code":"10101","plu":"10101","line-number":2},{"comment":[],"unit-price":0.0,"menu-item-code":"1.20001","qty":1,"inc-qty":0,"sub-unit-price":0.0,"tax-items":[],"default-qty":1,"item-type":"OPTION","sub-qty":0,"discount-applied":null,"added-qty":0,"product":"Bebida","item-discount":0.0,"dec-qty":0,"multiplied-qty":1.0,"chosen-qty":1,"price-key":null,"added-unit-price":0.0,"level":1,"item-price":0.0,"order-picture-id":6,"custom-properties":{},"part-code":"900000003","plu":"900000003","line-number":2},{"comment":[],"unit-price":2.5,"menu-item-code":"1.20001.900000003","qty":1,"inc-qty":0,"sub-unit-price":0.0,"tax-items":[{"base-amount-bd":2.5,"base-amount-ad":1.17,"tax-index":"PIS;P","tax-amount-bd":0.1,"tax-rule-id":1955574606,"tax-rate":0.04,"tax-amount-ad":0.0468},{"base-amount-bd":2.5,"base-amount-ad":1.17,"tax-index":"COFINS;C","tax-amount-bd":0.125,"tax-rule-id":2143042763,"tax-rate":0.05,"tax-amount-ad":0.0585},{"base-amount-bd":2.5,"base-amount-ad":1.17,"tax-index":"ICMS;I","tax-amount-bd":0.0,"tax-rule-id":1707847563,"tax-rate":0.13,"tax-amount-ad":0.0}],"default-qty":null,"item-type":"CANADD","sub-qty":null,"discount-applied":"600:1.33","added-qty":null,"product":"Coca Cola","item-discount":1.33,"dec-qty":0,"multiplied-qty":1.0,"chosen-qty":null,"price-key":"249108103238","added-unit-price":0.0,"level":2,"item-price":2.5,"order-picture-id":6,"custom-properties":{},"part-code":"2001","plu":"2001","line-number":2},{"comment":[],"unit-price":0.0,"menu-item-code":"1.20001","qty":1,"inc-qty":0,"sub-unit-price":0.0,"tax-items":[],"default-qty":1,"item-type":"OPTION","sub-qty":0,"discount-applied":null,"added-qty":0,"product":"Batata","item-discount":0.0,"dec-qty":0,"multiplied-qty":1.0,"chosen-qty":1,"price-key":null,"added-unit-price":0.0,"level":1,"item-price":0.0,"order-picture-id":6,"custom-properties":{},"part-code":"900000004","plu":"900000004","line-number":2},{"comment":[],"unit-price":2.5,"menu-item-code":"1.20001.900000004","qty":1,"inc-qty":0,"sub-unit-price":0.0,"tax-items":[{"base-amount-bd":2.5,"base-amount-ad":1.17,"tax-index":"PIS;P","tax-amount-bd":0.1,"tax-rule-id":1955574606,"tax-rate":0.04,"tax-amount-ad":0.0468},{"base-amount-bd":2.5,"base-amount-ad":1.17,"tax-index":"COFINS;C","tax-amount-bd":0.1,"tax-rule-id":2122209020,"tax-rate":0.04,"tax-amount-ad":0.0468},{"base-amount-bd":2.5,"base-amount-ad":1.17,"tax-index":"ICMS;I","tax-amount-bd":0.425,"tax-rule-id":1656883031,"tax-rate":0.17,"tax-amount-ad":0.1989}],"default-qty":null,"item-type":"CANADD","sub-qty":null,"discount-applied":"600:1.33","added-qty":null,"product":"Batata Média","item-discount":1.33,"dec-qty":0,"multiplied-qty":1.0,"chosen-qty":null,"price-key":"249108103246","added-unit-price":0.0,"level":2,"item-price":2.5,"order-picture-id":6,"custom-properties":{},"part-code":"3002","plu":"3002","line-number":2}],"total-gross":8.8,"store-code":"test-menu2","tax-applied":1.76,"xml-data":" ","total-after-discount":7.04,"session":"pos=1,user=1,count=1,period=20210707","sale-type-id":0,"tax-total":1.76,"due-amount":0.0,"change":0.0,"originator-code":"1","pos-code":"1","price-list":"EI","exemption":null,"discount-amount":10.0,"type-id":0,"price-basis":"G","tip":0.0,"business-dt":"2021-07-07","order-picture-tenders":[{"tender-type":"0","reference-amount":8.8,"tender-detail":null,"tip":null,"tender-amount":8.8,"order-picture-id":6,"change-amount":0.0}],"total-amount":17.04,"creation-dttm":"2021-07-07T17:13:39.855Z","total-tender":8.8,"pos-user-id":"1","total-gift":0,"discount-applied":null,"state-id":5,"order-discount-amount":0.0,"pod-type":"FL","order-picture-custom-order-properties":{"DISCOUNT_APPLY_OPERATOR":"1","BENEFIT_LIST":"[{\\"promotion_tender_amount\\": 0, \\"added_sale_lines\\": [], \\"benefit\\": \\"{\\\\\\"priority\\\\\\": 1, \\\\\\"userDescription\\\\\\": \\\\\\"Desconto por valor\\\\\\", \\\\\\"id\\\\\\": \\\\\\"1\\\\\\", \\\\\\"actions\\\\\\": {\\\\\\"1\\\\\\": {\\\\\\"discountAmountOnTotal\\\\\\": 10}}, \\\\\\"name\\\\\\": \\\\\\"Valor\\\\\\"}\\", \\"burnt_items\\": {}, \\"used_discount_ids\\": [600], \\"tender_id\\": \\"60\\"}]","FISCALIZATION_DATE":"2021-07-07T17:14:21","DISCOUNT_APPLY_AUTHORIZER":"1","DISCOUNT_APPLY_METHOD":"VLR11","DISCOUNT_APPLY_DATETIME":"2021-07-07T14:13:47"}}"""
        data += body
        data = data.encode("utf-8")

        event6 = Event("WsgiRequest", data=data, headers={"context": "localhost"})
        event7 = Event("ConsolidateSales", data=b"", headers={"context": "dev.backoffice.hmledp.com.br"})
        event8 = Event("FetchTefReconciliationReportData", data=b"", headers={"context": "127.0.0.1"})
        time.sleep(0.5)
        count = 1
        now = datetime.now()
        stop = now + timedelta(seconds=60)
        while datetime.now() < stop and count <= 1:
            pub_socket.send(format_event(event6))
            # pub_socket.send(format_event(event8))
            # pub_socket.send(format_event(event3))
            count += 1

        pub_socket.close()

        ctx.destroy()
        print(count)

    def test_message(self):
        context = zmq.Context()
        message_bus = ZeroMqMessageBus("EventManager", context)

        threads = []
        threads.extend(create_threads(1, send_slow_event, [message_bus, 1]))
        threads.extend(create_threads(30, send_fast_event, [message_bus, 5]))

        for t in threads:
            t.join()

        print("OK")


def create_threads(qty, target, args):
    threads = []
    for i in range(0, qty):
        t = Thread(target=target, args=args)
        t.start()
        threads.append(t)

    return threads


def send_slow_event(message_bus, qty):
    for i in range(0, qty):
        try:
            print(i)
            message_bus.publish_event(
                Event(
                    "WsgiRequest",
                    data=b"GET\0/test\0000\0000\0",
                    synchronous=True,
                    headers={"context": "localhost"}))
        except Exception as ex:
            print("Error on event: {}".format(ex))

    print("Sent {} events".format(qty))


def send_fast_event(message_bus, qty):
    for i in range(0, qty):
        reply = message_bus.publish_event(
            Event(
                "WsgiRequest",
                data=b"GET\0/test2\0000\0000\0",
                synchronous=True,
                headers={"context": "localhost"}))
        if reply.token != DefaultToken.TK_SYS_ACK:
            raise Exception("Error")
    print("Sent {} events".format(qty))


def convert_dict_to_bytes(d):
    if d is not None:
        ret = str(len(d)).encode("utf-8") + b"\0"
        for key in d.keys():
            ret += key.encode("utf-8") + b"\0" + d[key].encode("utf-8") + b"\0"
        ret = ret[:-1]
        return ret
    return b""


def format_event(event):
    # type: (Event) -> bytes
    type_data = event.event_type.encode("utf-8") if event.event_type is not None else b""
    event_data = b"" if event.data is None else event.data
    return event.subject.encode("utf-8") + b" " + type_data + b"\0" + convert_dict_to_bytes(event.headers) + b"\0" + event_data



class ConfigKey(object):
    LOCATOR_ADDRESS = "locatorAddress"
    PUB_ADDRESS = "pubAddress"
    SUB_ADDRESS = "subAddress"
    PULL_ADDRESS = "pullAddress"
    MIN_STREAMER_PORT = "minStreamerPort"
    MAX_STREAMER_PORT = "maxStreamerPort"
    PUBLIC_HOSTNAME = "publicHostname"
    STREAMER_POLL_TIMEOUT = "streamerPollTimeout"



from messagebus import NakMessage, MessageBus, Message, Event
from messageprocessorutil import UuidMessageProcessor
from zeromqmessagehandler import MessageDecoder

from .._EventManager import EventManager


class PublishProcessor(UuidMessageProcessor):

    def __init__(self, event_manager, event_name=None):
        # type: (EventManager, str) -> None
        super(PublishProcessor, self).__init__(event_name)
        self.event_manager = event_manager

    def parse_data(self, data):
        return MessageDecoder.parse_event(data)

    def call_business(self, event):
        # type: (Event) -> Message
        return self.event_manager.publish_sync_event(event)

    def format_response(self, message_bus, message, event, received_event, reply):
        # type: (MessageBus, Message, Event, Event, Message) -> None
        message_bus.reply_message(message, reply)

    def format_exception(self, message_bus, message, event, input_obj, exception):
        message_bus.reply_message(message, NakMessage(data=str(exception).encode("utf-8")))

    def format_parse_exception(self, message_bus, message, event, data, exception):
        message_bus.reply_message(message, NakMessage(data=str(exception).encode("utf-8")))



from messagebus import TokenCreator, TokenPriority

event_manager_group = "2"

TK_EVENT_MANAGER_UNSUBSCRIBE = TokenCreator.create_token(TokenPriority.low, event_manager_group, "2")
TK_EVENT_MANAGER_PUBLISH = TokenCreator.create_token(TokenPriority.low, event_manager_group, "3")
TK_EVENT_MANAGER_KEEP_ALIVE = TokenCreator.create_token(TokenPriority.low, event_manager_group, "4")
